
#include "ulib/graphics/lowlevel/img2raw.hh"
#include "ulib/graphics/lowlevel/yuv2rgb16.hh"

#if ENABLE_MMX
#include "ulib/graphics/lowlevel/yuv2rgb16mmx.hh"
#include "ulib/graphics/lowlevel/yuv2rgb32mmx.hh"
//#include "ulib/graphics/lowlevel/yuv2rgb_mmx.hh"
#include "ulib/graphics/lowlevel/grey2rgb32mmx.hh"
#endif

#include <iostream.h>


/* Place RGB components in 32bit entities in this memory order: - R G B independent of endianness issues. */
class i2r_xrgb : public Image2Raw_TransformRGB
{
public:
  virtual ~i2r_xrgb() { }
  virtual void Transform(const Image_RGB<Pixel>&,uint8* mem);

  virtual const char* TransformationName() { return "scalar XRGB"; }
};

/* Place RGB components in 32bit entities in this memory order: - B G R independent of endianness issues. */
class i2r_xbgr : public Image2Raw_TransformRGB
{
public:
  virtual ~i2r_xbgr() { }
  virtual void Transform(const Image_RGB<Pixel>&,uint8* mem);

  virtual const char* TransformationName() { return "scalar XBGR"; }
};

/* Place RGB components in 32bit entities in arbitrary order.
   For the special cases that can be handled by the classes i2r_xrgb and i2r_xbgr you
   should use them as they are a bit faster.
 */
class i2r_32bit : public Image2Raw_TransformRGB
{
public:
  virtual ~i2r_32bit() { }
  virtual void Transform(const Image_RGB<Pixel>&,uint8* mem);

  enum Channel { Red=0, Green=1, Blue=2, Empty=3 };

  void SetPixelSequence(Channel p[4]) { c1=p[0]; c2=p[1]; c3=p[2]; c4=p[3]; }

  virtual const char* TransformationName() { return "scalar 32bit RGB"; }

private:
  Channel c1,c2,c3,c4;
};

/* Convert YUV 4:2:0 and place RGB components in 32bit entities in arbitrary order.
 */
class i2r_yuv_32bit : public Image2Raw_TransformYUV
{
public:
  virtual ~i2r_yuv_32bit() { }
  virtual void Transform(const Image_YUV<Pixel>&,uint8* mem);

  enum Channel { Red=0, Green=1, Blue=2 };

  void SetChannelPositions(int r,int g,int b) { rpos=r; gpos=g; bpos=b; }

  virtual const char* TransformationName() { return "scalar YUV to 32bit RGB"; }

private:
  int rpos,gpos,bpos;
};

/* Convert greyscale to RGB components in 32bit entities in arbitrary order.
 */
class i2r_grey_32bit : public Image2Raw_TransformYUV
{
public:
  virtual ~i2r_grey_32bit() { }
  virtual void Transform(const Image_YUV<Pixel>&,uint8* mem);

  virtual const char* TransformationName() { return "scalar grey to 32bit RGB"; }
};



void CalcBitsShift(uint32 mask,int& bits,int& shift)
{
  assert(mask!=0);

  shift=0;
  while ((mask&1)==0) { shift++; mask>>=1; }
  bits=0;
  while (mask&1) { bits++; mask>>=1; }

  assert(mask==0); // This may fail if there are more than one continuous sequences if ones (like 000111000011100).
}

Image2Raw::Image2Raw()
  : d_mem(NULL),
    d_rgbinput(false),
    d_chroma(Chroma420),
    d_greyscale(false),
    d_rgbtransform(NULL)
{
}

Image2Raw::~Image2Raw()
{
  if (d_rgbinput) { if (d_rgbtransform) delete d_rgbtransform; }
  else            { if (d_yuvtransform) delete d_yuvtransform; }
}

void Image2Raw::SetOutputMem(uint8* mem)
{
  d_mem=mem;
}

void Image2Raw::SetOutputParam(int bytes_per_line,int bits_per_pixel,bool littleendian,
			       uint32 redmask,uint32 greenmask,uint32 bluemask)
{
  CalcBitsShift(redmask  ,red_bits  ,red_shift  ); d_redmask  =redmask;
  CalcBitsShift(greenmask,green_bits,green_shift); d_greenmask=greenmask;
  CalcBitsShift(bluemask ,blue_bits ,blue_shift ); d_bluemask =bluemask;

  d_bytes_per_line = bytes_per_line;
  d_bits_per_pixel = bits_per_pixel;
  d_littleendian = littleendian;

  bool transformation_found = false;

  if (d_rgbinput)
    {
      if (red_bits==8 && green_bits==8 && blue_bits==8)
	{
	  // 24 bit / 32 bit modes

	  if ((!littleendian && red_shift==16 && green_shift== 8 && blue_shift== 0) ||
	      ( littleendian && red_shift== 8 && green_shift==16 && blue_shift==24))
	    {
	      d_rgbtransform = new i2r_xrgb;
	      transformation_found = true;
	    }
	  else
	    if ((!littleendian && red_shift== 0 && green_shift== 8 && blue_shift==16) ||
		( littleendian && red_shift==24 && green_shift==16 && blue_shift== 8))
	      {
		d_rgbtransform = new i2r_xbgr;
		transformation_found = true;
	      }
	    else
	      if ((red_shift%8)==0 && (green_shift%8)==0 && (blue_shift%8)==0 && d_bits_per_pixel==32)
		{
		  i2r_32bit* transform = new i2r_32bit;
		  d_rgbtransform = transform;
		  i2r_32bit::Channel c[4];
		  c[0]=c[1]=c[2]=c[3]=i2r_32bit::Empty;
		  c[(24-red_shift  )/8]=i2r_32bit::Red;
		  c[(24-green_shift)/8]=i2r_32bit::Green;
		  c[(24-blue_shift )/8]=i2r_32bit::Blue;

		  if (littleendian)
		    { swap(c[0],c[3]); swap(c[1],c[2]); }

		  transform->SetPixelSequence(c);
		  transformation_found = true;
		}
	}
    }
  else
    {
      if (d_greyscale)
	{
#if ENABLE_MMX
	  i2r_grey_32bit_mmx* transform = new i2r_grey_32bit_mmx;
#else
	  i2r_grey_32bit* transform = new i2r_grey_32bit;
#endif
	  d_yuvtransform = transform;

	  transformation_found = true;
	}
      else
	{
	  if (d_chroma==Chroma420)
	    {
	      if ((red_shift%8)==0 && (green_shift%8)==0 && (blue_shift%8)==0 && d_bits_per_pixel==32)
		{
		  int r=3-red_shift  /8;
		  int g=3-green_shift/8;
		  int b=3-blue_shift /8;

		  if (littleendian)
		    { r=3-r; g=3-g; b=3-b; }

#if ENABLE_MMX
		  if (r==2 && g==1 && b==0)
		    {
		      i2r_32bit_BGR_mmx* transform = new i2r_32bit_BGR_mmx;
		      //i2r_32bit_BGR_mmx2* transform = new i2r_32bit_BGR_mmx2;
		      d_yuvtransform = transform;
		      
		      transformation_found = true;
		    }
		  else
#endif
		    {
		      i2r_yuv_32bit* transform = new i2r_yuv_32bit;
		      d_yuvtransform = transform;
		      
		      transform->SetChannelPositions(r,g,b);
		      transformation_found = true;
		    }
		}
	      else if (bits_per_pixel==16)
		{
#if ENABLE_MMX
		  if (littleendian &&
		      red_bits  >=4 && red_bits  <=6 &&
		      green_bits>=4 && green_bits<=6 &&
		      blue_bits >=4 && blue_bits <=6 &&
		      blue_shift==0 &&
		      red_shift>=8)
		    {
		      i2r_16bit_mmx* transform = new i2r_16bit_mmx;
		      d_yuvtransform = transform;

		      transform->SetMask(redmask,greenmask,bluemask);
		      transformation_found = true;
		    }
		  else
#endif
		    {
		      i2r_16bit* transform = new i2r_16bit;
		      d_yuvtransform = transform;

		      transform->SetMask(redmask,greenmask,bluemask);
		      transform->SetLittleEndian(littleendian);
		      transformation_found = true;
		    }
		}
	    }
	  else
	    {
	    }
	}
    }

  assert(transformation_found);

#if 0
  cout << "Display transformation: " << (d_rgbinput ?
                                         d_rgbtransform->TransformationName() :
                                         d_yuvtransform->TransformationName()) << endl;
#endif

  if (d_rgbinput) d_rgbtransform->SetBytesPerLine(d_bytes_per_line);
  else            d_yuvtransform->SetBytesPerLine(d_bytes_per_line);
}

void Image2Raw::TransformRGB(const Image_RGB<Pixel>& img)
{
  assert(d_rgbtransform);
  d_rgbtransform->Transform(img,d_mem);
}

void Image2Raw::TransformYUV(const Image_YUV<Pixel>& img)
{
  assert(d_yuvtransform);
  d_yuvtransform->Transform(img,d_mem);
}








static int* clip_0_255;
static int s_clip[1024];


static struct InitClip
{
  InitClip()
    {
      clip_0_255 = &s_clip[512];

      for (int i=-512;i<512;i++)
        {
          if (i<0)
            {
              clip_0_255[i]=0;
            }
          else if (i>255)
            {
              clip_0_255[i]=255;
            }
          else
            {
              clip_0_255[i]=i;
            }
        }
    }
} dummy_23874678;




void i2r_xrgb::Transform(const Image_RGB<Pixel>& img,uint8* mem)
{
  ImageParam param;
  img.GetParam(param);

  const Pixel*const* r = img.AskFrameR_const();
  const Pixel*const* g = img.AskFrameG_const();
  const Pixel*const* b = img.AskFrameB_const();

  for (int y=0;y<param.height;y++)
    {
      uint8* p = mem+y*d_bpl;

      for (int x=0;x<param.width;x++)
	{
	  p[1] = r[y][x];
	  p[2] = g[y][x];
	  p[3] = b[y][x];
	  p+=4;
	}
    }
}

void i2r_xbgr::Transform(const Image_RGB<Pixel>& img,uint8* mem)
{
  ImageParam param;
  img.GetParam(param);

  const Pixel*const* r = img.AskFrameR_const();
  const Pixel*const* g = img.AskFrameG_const();
  const Pixel*const* b = img.AskFrameB_const();

  for (int y=0;y<param.height;y++)
    {
      uint8* p = mem+y*d_bpl;

      for (int x=0;x<param.width;x++)
	{
	  p[1] = b[y][x];
	  p[2] = g[y][x];
	  p[3] = r[y][x];
	  p+=4;
	}
    }
}

void i2r_32bit::Transform(const Image_RGB<Pixel>& img,uint8* mem)
{
  ImageParam param;
  img.GetParam(param);

  const Pixel*const* c[4];
  c[0] = img.AskFrameR_const();
  c[1] = img.AskFrameG_const();
  c[2] = img.AskFrameB_const();
  c[3] = img.AskFrameR_const(); // don't care

  for (int y=0;y<param.height;y++)
    {
      uint8* p = mem+y*d_bpl;

      for (int x=0;x<param.width;x++)
	{
	  p[0] = c[c1][y][x];
	  p[1] = c[c2][y][x];
	  p[2] = c[c3][y][x];
	  p[3] = c[c4][y][x];
	  p+=4;
	}
    }
}


void i2r_yuv_32bit::Transform(const Image_YUV<Pixel>& img,uint8* mem)
{
  ImageParam_YUV param;
  img.GetParam(param);

  assert(param.chroma==Chroma420);

  const Pixel*const * pix_y = img.AskFrameY_const();
  const Pixel*const * pix_u = img.AskFrameU_const();
  const Pixel*const * pix_v = img.AskFrameV_const();

  for (int y=0;y<param.height;y+=2)
    {
      uint8* membuf8a = mem + d_bpl*y;
      uint8* membuf8b = mem + d_bpl*y +d_bpl;

      const Pixel* cbp = pix_u[y/2];
      const Pixel* crp = pix_v[y/2];
      const Pixel* y0p = pix_y[y  ];
      const Pixel* y1p = pix_y[y+1];

      for (int x=0;x<param.width;x+=2)
        {
          int cb=((int)*cbp++) -128;
          int cr=((int)*crp++) -128;

          int r0 = (int)(          + 409*cr);
          int g0 = (int)( - 100*cb - 208*cr);
          int b0 = (int)( + 516*cb         );
          
          int yy=(((int)*y0p++) -16)*298;
          int red   = (r0+yy)>>8; red  = clip_0_255[red];
          int green = (g0+yy)>>8; green= clip_0_255[green];
          int blue  = (b0+yy)>>8; blue = clip_0_255[blue];
          membuf8a[rpos] = red;
          membuf8a[gpos] = green;
          membuf8a[bpos] = blue;
          membuf8a+=4;

          yy=(((int)*y0p++) -16)*298;
          red   = (r0+yy)>>8; red  = clip_0_255[red];
          green = (g0+yy)>>8; green= clip_0_255[green];
          blue  = (b0+yy)>>8; blue = clip_0_255[blue];
          membuf8a[rpos] = red;
          membuf8a[gpos] = green;
          membuf8a[bpos] = blue;
          membuf8a+=4;
          
          yy=(((int)*y1p++) -16)*298;
          red   = (r0+yy)>>8; red  = clip_0_255[red];
          green = (g0+yy)>>8; green= clip_0_255[green];
          blue  = (b0+yy)>>8; blue = clip_0_255[blue];
          membuf8b[rpos] = red;
          membuf8b[gpos] = green;
          membuf8b[bpos] = blue;
          membuf8b+=4;
          
          yy=(((int)*y1p++) -16)*298;
          red   = (r0+yy)>>8; red  = clip_0_255[red];
          green = (g0+yy)>>8; green= clip_0_255[green];
          blue  = (b0+yy)>>8; blue = clip_0_255[blue];
          membuf8b[rpos] = red;
          membuf8b[gpos] = green;
          membuf8b[bpos] = blue;
          membuf8b+=4;
        }
    }
}



void i2r_grey_32bit::Transform(const Image_YUV<Pixel>& img,uint8* mem)
{
  ImageParam_YUV param;
  img.GetParam(param);

  const Pixel*const * pix_y = img.AskFrameY_const();

  for (int y=0;y<param.height;y++)
    {
      uint32* membuf = (uint32*)(mem + d_bpl*y);

      for (int x=0;x<param.width;x++)
        {
	  uint32 val = pix_y[y][x];
	  val |= val<<16;
	  val |= val<< 8;

          *membuf++ = val;
        }
    }
}
